@@ -1,20 +1,26 @@
 # This file is part of Xpra.
-# Copyright (C) 2018-2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2018-2024 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
-#pylint: disable-msg=E1101
+# pylint: disable-msg=E1101
 
-import os
 from time import monotonic
 from threading import RLock
+from typing import Any
+from collections.abc import Sequence
 
 from xpra.log import Logger
 from xpra.scripts.config import FALSE_OPTIONS
 from xpra.net import compression
-from xpra.os_util import OSEnvContext, WIN32
-from xpra.util import envint, envbool, csv, typedict, XPRA_WEBCAM_NOTIFICATION_ID
-from xpra.client.mixins.stub_client_mixin import StubClientMixin
+from xpra.net.common import PacketType
+from xpra.os_util import WIN32, gi_import
+from xpra.util.objects import typedict
+from xpra.util.str_fn import csv
+from xpra.util.env import envint, envbool, OSEnvContext
+from xpra.common import NotificationID
+from xpra.client.base.stub_client_mixin import StubClientMixin
 
+GLib = gi_import("GLib")
 
 log = Logger("webcam")
 
@@ -31,42 +37,34 @@ class WebcamForwarder(StubClientMixin):
 
     def __init__(self):
         super().__init__()
-        #webcam:
+        # webcam:
         self.webcam_option = ""
         self.webcam_forwarding = False
         self.webcam_device = None
         self.webcam_device_no = -1
         self.webcam_last_ack = -1
-        self.webcam_ack_check_timer = None
-        self.webcam_send_timer = None
+        self.webcam_ack_check_timer = 0
+        self.webcam_send_timer = 0
         self.webcam_lock = RLock()
         self.server_webcam = False
+        self.server_webcam_encodings: Sequence[str] = ()
         self.server_virtual_video_devices = 0
-        if not hasattr(self, "send"):
-            def nosend(*_args):
-                """
-                pretend to have a send method for testing
-                """
-            self.send = nosend
-        #duplicated from encodings mixin:
-        self.server_encodings = []
+        # duplicated from encodings mixin:
+        self.server_encodings: Sequence[str] = ()
         if not hasattr(self, "server_ping_latency"):
             from collections import deque
-            self.server_ping_latency = deque(maxlen=1000)
+            self.server_ping_latency: deque[tuple[float, float]] = deque(maxlen=1000)
 
-    def cleanup(self):
+    def cleanup(self) -> None:
         self.stop_sending_webcam()
 
-
-    def init(self, opts):
+    def init(self, opts) -> None:
         self.webcam_option = opts.webcam
         self.webcam_forwarding = self.webcam_option.lower() not in FALSE_OPTIONS
         self.server_webcam = False
         self.server_virtual_video_devices = 0
         if self.webcam_forwarding:
-            with OSEnvContext():
-                os.environ["LANG"] = "C"
-                os.environ["LC_ALL"] = "C"
+            with OSEnvContext(LANG="C", LC_ALL="C"):
                 try:
                     import cv2
                     from PIL import Image
@@ -80,35 +78,39 @@ class WebcamForwarder(StubClientMixin):
                     self.webcam_forwarding = False
         log("webcam forwarding: %s", self.webcam_forwarding)
 
-
-    def get_caps(self) -> dict:
+    def get_caps(self) -> dict[str, Any]:
         if not self.webcam_forwarding:
             return {}
-        return {"webcam" : True}
+        return {"webcam": True}
 
-
-    def parse_server_capabilities(self, c : typedict) -> bool:
-        self.server_webcam = c.boolget("webcam")
-        self.server_webcam_encodings = c.strtupleget("webcam.encodings", ("png", "jpeg"))
-        self.server_virtual_video_devices = c.intget("virtual-video-devices")
+    def parse_server_capabilities(self, c: typedict) -> bool:
+        v = c.get("webcam")
+        if isinstance(v, dict):
+            cdict = typedict(v)
+            self.server_webcam = cdict.boolget("enabled")
+            self.server_webcam_encodings = cdict.strtupleget("encodings", ("png", "jpeg"))
+            self.server_virtual_video_devices = cdict.intget("devices")
+        else:
+            # pre v6 / 5.0.2
+            self.server_webcam = c.boolget("webcam")
+            self.server_webcam_encodings = c.strtupleget("webcam.encodings", ("png", "jpeg"))
+            self.server_virtual_video_devices = c.intget("virtual-video-devices")
         log("webcam server support: %s (%i devices, encodings: %s)",
             self.server_webcam, self.server_virtual_video_devices, csv(self.server_webcam_encodings))
-        if self.webcam_forwarding and self.server_webcam and self.server_virtual_video_devices>0:
-            if self.webcam_option=="on" or self.webcam_option.find("/dev/video")>=0:
+        if self.webcam_forwarding and self.server_webcam and self.server_virtual_video_devices > 0:
+            if self.webcam_option == "on" or self.webcam_option.find("/dev/video") >= 0:
                 self.start_sending_webcam()
         return True
 
-
-    def webcam_state_changed(self):
-        self.idle_add(self.emit, "webcam-changed")
-
+    def webcam_state_changed(self) -> None:
+        GLib.idle_add(self.emit, "webcam-changed")
 
     ######################################################################
-    def start_sending_webcam(self):
+    def start_sending_webcam(self) -> None:
         with self.webcam_lock:
             self.do_start_sending_webcam(self.webcam_option)
 
-    def do_start_sending_webcam(self, device_str):
+    def do_start_sending_webcam(self, device_str) -> None:
         self.show_progress(100, "forwarding webcam")
         assert self.server_webcam
         device = 0
@@ -116,8 +118,8 @@ class WebcamForwarder(StubClientMixin):
         try:
             from xpra.platform.webcam import get_virtual_video_devices, get_all_video_devices
             virt_devices = get_virtual_video_devices()
-            all_video_devices = get_all_video_devices()     #pylint: disable=assignment-from-none
-            non_virtual = dict((k,v) for k,v in all_video_devices.items() if k not in virt_devices)
+            all_video_devices = get_all_video_devices()  # pylint: disable=assignment-from-none
+            non_virtual = {k: v for k, v in all_video_devices.items() if k not in virt_devices}
             log("virtual video devices=%s", virt_devices)
             log("all_video_devices=%s", all_video_devices)
             log("found %s known non-virtual video devices: %s", len(non_virtual), non_virtual)
@@ -133,10 +135,10 @@ class WebcamForwarder(StubClientMixin):
                 device = int(device_str)
             except ValueError:
                 p = device_str.find("video")
-                if p>=0:
+                if p >= 0:
                     try:
                         log("device_str: %s", device_str[p:])
-                        device = int(device_str[p+len("video"):])
+                        device = int(device_str[p + len("video"):])
                     except ValueError:
                         device = 0
         if device in virt_devices:
@@ -151,54 +153,54 @@ class WebcamForwarder(StubClientMixin):
         log("do_start_sending_webcam(%s) device=%i", device_str, device)
         self.webcam_frame_no = 0
         try:
-            #test capture:
-            webcam_device = cv2.VideoCapture(device)        #0 -> /dev/video0 @UndefinedVariable
+            # test capture:
+            webcam_device = cv2.VideoCapture(device)  # 0 -> /dev/video0 @UndefinedVariable
             ret, frame = webcam_device.read()
             log("test capture using %s: %s, %s", webcam_device, ret, frame is not None)
             assert ret, "no device or permission"
             assert frame is not None, "no data"
-            assert frame.ndim==3, "unexpected  number of dimensions: %s" % frame.ndim
-            w, h, Bpp = frame.shape
-            assert Bpp==3, "unexpected number of bytes per pixel: %s" % Bpp
-            assert frame.size==w*h*Bpp
+            assert frame.ndim == 3, "unexpected  number of dimensions: %s" % frame.ndim
+            h, w, Bpp = frame.shape
+            assert Bpp == 3, "unexpected number of bytes per pixel: %s" % Bpp
+            assert frame.size == w * h * Bpp
             self.webcam_device_no = device
             self.webcam_device = webcam_device
             self.send("webcam-start", device, w, h)
             self.webcam_state_changed()
             log("webcam started")
             if self.send_webcam_frame():
-                delay = 1000//WEBCAM_TARGET_FPS
+                delay = 1000 // WEBCAM_TARGET_FPS
                 log("webcam timer with delay=%ims for %i fps target)", delay, WEBCAM_TARGET_FPS)
                 self.cancel_webcam_send_timer()
-                self.webcam_send_timer = self.timeout_add(delay, self.may_send_webcam_frame)
+                self.webcam_send_timer = GLib.timeout_add(delay, self.may_send_webcam_frame)
         except Exception as e:
             log.warn("webcam test capture failed: %s", e)
 
-    def cancel_webcam_send_timer(self):
+    def cancel_webcam_send_timer(self) -> None:
         wst = self.webcam_send_timer
         if wst:
-            self.webcam_send_timer = None
-            self.source_remove(wst)
+            self.webcam_send_timer = 0
+            GLib.source_remove(wst)
 
-    def cancel_webcam_check_ack_timer(self):
+    def cancel_webcam_check_ack_timer(self) -> None:
         wact = self.webcam_ack_check_timer
         if wact:
-            self.webcam_ack_check_timer = None
-            self.source_remove(wact)
+            self.webcam_ack_check_timer = 0
+            GLib.source_remove(wact)
 
-    def webcam_check_acks(self, ack=0):
-        self.webcam_ack_check_timer = None
+    def webcam_check_acks(self, ack=0) -> None:
+        self.webcam_ack_check_timer = 0
         log("check_acks: webcam_last_ack=%s", self.webcam_last_ack)
-        if self.webcam_last_ack<ack:
+        if self.webcam_last_ack < ack:
             log.warn("Warning: no acknowledgements received from the server for frame %i, stopping webcam", ack)
             self.stop_sending_webcam()
 
-    def stop_sending_webcam(self):
+    def stop_sending_webcam(self) -> None:
         log("stop_sending_webcam()")
         with self.webcam_lock:
             self.do_stop_sending_webcam()
 
-    def do_stop_sending_webcam(self):
+    def do_stop_sending_webcam(self) -> None:
         self.cancel_webcam_send_timer()
         self.cancel_webcam_check_ack_timer()
         wd = self.webcam_device
@@ -217,31 +219,31 @@ class WebcamForwarder(StubClientMixin):
             log.error("Error closing webcam device %s: %s", wd, e)
         self.webcam_state_changed()
 
-    def may_send_webcam_frame(self):
-        self.webcam_send_timer = None
-        if self.webcam_device_no<0 or not self.webcam_device:
+    def may_send_webcam_frame(self) -> bool:
+        self.webcam_send_timer = 0
+        if self.webcam_device_no < 0 or not self.webcam_device:
             return False
-        not_acked = self.webcam_frame_no-1-self.webcam_last_ack
-        #not all frames have been acked
+        not_acked = self.webcam_frame_no - 1 - self.webcam_last_ack
+        # not all frames have been acked
         latency = 100
-        spl = tuple(x for _,x in self.server_ping_latency)
+        spl = tuple(x for _, x in self.server_ping_latency)
         if spl:
             latency = int(1000 * sum(spl) / len(spl))
-        #how many frames should be in flight
-        n = max(1, latency // (1000//WEBCAM_TARGET_FPS))    #20fps -> 50ms target between frames
-        if not_acked>0 and not_acked>n:
+        # how many frames should be in flight
+        n = max(1, latency // (1000 // WEBCAM_TARGET_FPS))  # 20fps -> 50ms target between frames
+        if not_acked > 0 and not_acked > n:
             log("may_send_webcam_frame() latency=%i, not acked=%i, target=%i - will wait for next ack",
                 latency, not_acked, n)
             return False
         log("may_send_webcam_frame() latency=%i, not acked=%i, target=%i - trying to send now", latency, not_acked, n)
         return self.send_webcam_frame()
 
-    def send_webcam_frame(self):
+    def send_webcam_frame(self) -> bool:
         if not self.webcam_lock.acquire(False):
             return False
         log("send_webcam_frame() webcam_device=%s", self.webcam_device)
         try:
-            assert self.webcam_device_no>=0, "device number is not set"
+            assert self.webcam_device_no >= 0, "device number is not set"
             assert self.webcam_device, "no webcam device to capture from"
             from xpra.codecs.pillow.encoder import get_encodings
             client_webcam_encodings = get_encodings()
@@ -261,14 +263,14 @@ class WebcamForwarder(StubClientMixin):
             import cv2
             ret, frame = self.webcam_device.read()
             assert ret, "capture failed"
-            assert frame.ndim==3, "invalid frame data"
+            assert frame.ndim == 3, "invalid frame data"
             h, w, Bpp = frame.shape
-            assert Bpp==3 and frame.size==w*h*Bpp
+            assert Bpp == 3 and frame.size == w * h * Bpp
             rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)  # @UndefinedVariable
             end = monotonic()
-            log("webcam frame capture took %ims", (end-start)*1000)
+            log("webcam frame capture took %ims", (end - start) * 1000)
             start = monotonic()
-            from PIL import Image
+            from PIL import Image  # @UnresolvedImport
             from io import BytesIO
             image = Image.fromarray(rgb)
             buf = BytesIO()
@@ -276,13 +278,13 @@ class WebcamForwarder(StubClientMixin):
             data = buf.getvalue()
             buf.close()
             end = monotonic()
-            log("webcam frame compression to %s took %ims", encoding, (end-start)*1000)
+            log("webcam frame compression to %s took %ims", encoding, (end - start) * 1000)
             frame_no = self.webcam_frame_no
             self.webcam_frame_no += 1
             self.send("webcam-frame", self.webcam_device_no, frame_no, encoding,
                       w, h, compression.Compressed(encoding, data))
             self.cancel_webcam_check_ack_timer()
-            self.webcam_ack_check_timer = self.timeout_add(10*1000, self.webcam_check_acks)
+            self.webcam_ack_check_timer = GLib.timeout_add(10 * 1000, self.webcam_check_acks)
             return True
         except Exception as e:
             log.error("webcam frame %i failed", self.webcam_frame_no, exc_info=True)
@@ -290,22 +292,21 @@ class WebcamForwarder(StubClientMixin):
             self.stop_sending_webcam()
             summary = "Webcam forwarding has failed"
             body = "The system encountered the following error:\n" + \
-                ("%s\n" % e)
-            self.may_notify(XPRA_WEBCAM_NOTIFICATION_ID, summary, body, expire_timeout=10*1000, icon_name="webcam")
+                   ("%s\n" % e)
+            self.may_notify(NotificationID.WEBCAM, summary, body, expire_timeout=10 * 1000, icon_name="webcam")
             return False
         finally:
             self.webcam_lock.release()
 
-
     ######################################################################
-    #packet handlers
-    def _process_webcam_stop(self, packet):
+    # packet handlers
+    def _process_webcam_stop(self, packet: PacketType) -> None:
         device_no = packet[1]
-        if device_no!=self.webcam_device_no:
+        if device_no != self.webcam_device_no:
             return
         self.stop_sending_webcam()
 
-    def _process_webcam_ack(self, packet):
+    def _process_webcam_ack(self, packet: PacketType) -> None:
         log("process_webcam_ack: %s", packet)
         with self.webcam_lock:
             if self.webcam_device:
@@ -313,12 +314,11 @@ class WebcamForwarder(StubClientMixin):
                 self.webcam_last_ack = frame_no
                 if self.may_send_webcam_frame():
                     self.cancel_webcam_send_timer()
-                    delay = 1000//WEBCAM_TARGET_FPS
+                    delay = 1000 // WEBCAM_TARGET_FPS
                     log("new webcam timer with delay=%ims for %i fps target)", delay, WEBCAM_TARGET_FPS)
-                    self.webcam_send_timer = self.timeout_add(delay, self.may_send_webcam_frame)
+                    self.webcam_send_timer = GLib.timeout_add(delay, self.may_send_webcam_frame)
 
-
-    def init_authenticated_packet_handlers(self):
+    def init_authenticated_packet_handlers(self) -> None:
         log("init_authenticated_packet_handlers()")
         self.add_packet_handler("webcam-stop", self._process_webcam_stop)
         self.add_packet_handler("webcam-ack", self._process_webcam_ack)
