@@ -5,13 +5,18 @@
 # later version. See the file COPYING for details.
 
 import os
+import sys
 import signal
+from typing import Any
 from queue import SimpleQueue
+from collections.abc import Sequence
 from multiprocessing import Process
 
 from xpra.server.proxy.instance_base import ProxyInstance
 from xpra.server.proxy.queue_scheduler import QueueScheduler
 from xpra.server.util import setuidgid
+from xpra.server.mixins.control import ControlHandler
+from xpra.server import features
 from xpra.scripts.server import deadly_signal
 from xpra.net.protocol.factory import get_client_protocol_class, get_server_protocol_class
 from xpra.net.protocol.constants import CONNECTION_LOST
@@ -19,8 +24,9 @@ from xpra.net.protocol.socket_handler import SocketProtocol
 from xpra.net.socket_util import SOCKET_DIR_MODE, create_unix_domain_socket, handle_socket_error
 from xpra.net.bytestreams import SocketConnection, SOCKET_TIMEOUT
 from xpra.net.common import PacketType
-from xpra.os_util import POSIX, getuid, getgid, get_username_for_uid, gi_import
+from xpra.os_util import POSIX, getuid, getgid, get_username_for_uid, get_machine_id
 from xpra.util.env import osexpand
+from xpra.exit_codes import ExitValue, ExitCode
 from xpra.scripts.config import str_to_bool
 from xpra.util.system import SIGNAMES, register_SIGUSR_signals, set_proc_title
 from xpra.util.objects import typedict
@@ -32,8 +38,6 @@ from xpra.common import ConnectionMessage, SocketState, noerr, FULL_INFO
 from xpra.platform.dotxpra import DotXpra
 from xpra.log import Logger
 
-GLib = gi_import("GLib")
-
 log = Logger("proxy")
 enclog = Logger("encoding")
 
@@ -52,16 +56,18 @@ def set_blocking(conn) -> None:
         log("cannot set %s to blocking mode", conn)
 
 
-class ProxyInstanceProcess(ProxyInstance, QueueScheduler, Process):
+class ProxyInstanceProcess(ProxyInstance, QueueScheduler, ControlHandler, Process):
 
-    def __init__(self, uid, gid, env_options, session_options, socket_dir,
-                 video_encoder_modules, pings,
-                 client_conn, disp_desc, client_state,
-                 cipher, cipher_mode, encryption_key, server_conn, caps, message_queue):
+    def __init__(self, uid: int, gid: int, env_options: dict[str, str], session_options: dict[str, str],
+                 socket_dir: str,
+                 video_encoder_modules: Sequence[str], pings: int,
+                 client_conn, disp_desc: dict[str, Any], client_state: dict[str, Any],
+                 cipher: str, cipher_mode: str, encryption_key: bytes, server_conn, caps: typedict, message_queue):
         ProxyInstance.__init__(self, session_options,
                                video_encoder_modules, pings,
                                disp_desc, cipher, cipher_mode, encryption_key, caps)
         QueueScheduler.__init__(self)
+        ControlHandler.__init__(self)
         Process.__init__(self, name=str(client_conn), daemon=False)
         self.client_conn = client_conn
         self.server_conn = server_conn
@@ -121,8 +127,9 @@ class ProxyInstanceProcess(ProxyInstance, QueueScheduler, Process):
 
     ################################################################################
 
-    def run(self) -> None:
-        register_SIGUSR_signals(GLib.idle_add)
+    def run(self) -> ExitValue:
+        register_SIGUSR_signals(self.idle_add)
+        ControlHandler.add_default_control_commands(self, features.control)
         client_protocol_class = get_client_protocol_class(self.client_conn.socktype)
         server_protocol_class = get_server_protocol_class(self.server_conn.socktype)
         self.client_protocol = client_protocol_class(self.client_conn,
@@ -164,7 +171,7 @@ class ProxyInstanceProcess(ProxyInstance, QueueScheduler, Process):
         start_thread(self.server_message_queue, "server message queue")
 
         if not self.create_control_socket():
-            return
+            return ExitCode.FAILURE
         self.control_socket_thread = start_thread(self.control_socket_loop, "control", daemon=True)
 
         self.main_queue = SimpleQueue()
@@ -175,8 +182,10 @@ class ProxyInstanceProcess(ProxyInstance, QueueScheduler, Process):
             QueueScheduler.run(self)
         except KeyboardInterrupt as e:
             self.stop(None, str(e))
+            return 128 + int(signal.SIGINT)
         finally:
             log("ProxyProcess.run() ending %s", os.getpid())
+        return 0
 
     def start_network_threads(self) -> None:
         log("start_network_threads()")
@@ -189,7 +198,7 @@ class ProxyInstanceProcess(ProxyInstance, QueueScheduler, Process):
     def create_control_socket(self) -> bool:
         assert self.socket_dir
 
-        def stop(msg):
+        def stop(msg) -> None:
             self.stop(None, f"cannot create the proxy control socket: {msg}")
 
         username = get_username_for_uid(self.uid)
@@ -269,12 +278,12 @@ class ProxyInstanceProcess(ProxyInstance, QueueScheduler, Process):
         sc = SocketConnection(sock, sockname, address, target, "socket")
         log.info("New proxy instance control connection received:")
         log.info(" '%s'", sc)
-        protocol = SocketProtocol(sc, self.process_control_packet)
+        protocol = SocketProtocol(sc, self.process_control_packet, scheduler=self)
         protocol.large_packets.append("info-response")
         self.potential_protocols.append(protocol)
         protocol.enable_default_encoder()
         protocol.start()
-        GLib.timeout_add(SOCKET_TIMEOUT * 1000, self.verify_connection_accepted, protocol)
+        self.timeout_add(SOCKET_TIMEOUT * 1000, self.verify_connection_accepted, protocol)
 
     def verify_connection_accepted(self, protocol) -> None:
         if not protocol.is_closed() and protocol in self.potential_protocols:
@@ -329,7 +338,7 @@ class ProxyInstanceProcess(ProxyInstance, QueueScheduler, Process):
             else:
                 info = {"error": "`info` requests are not enabled for this connection"}
             proto.send_now(("hello", info))
-            GLib.timeout_add(5 * 1000, self.send_disconnect, proto, ConnectionMessage.CLIENT_EXIT_TIMEOUT,
+            self.timeout_add(5 * 1000, self.send_disconnect, proto, ConnectionMessage.CLIENT_EXIT_TIMEOUT,
                              "info sent")
             return True
         if request == "stop":
@@ -338,16 +347,41 @@ class ProxyInstanceProcess(ProxyInstance, QueueScheduler, Process):
             else:
                 log.warn("Warning: `stop` requests are not allowed for this connection")
             return True
+        if request == "command":
+            command_req = tuple(str(x) for x in caps.tupleget("command_request"))
+            self.idle_add(self.handle_command_request, proto, *command_req)
+            return True
         if request == "version":
             version = XPRA_VERSION
             if caps.boolget("full-version-request") and FULL_INFO:
                 version = full_version_str()
             proto.send_now(("hello", {"version": version}))
-            GLib.timeout_add(5 * 1000, self.send_disconnect, proto, ConnectionMessage.CLIENT_EXIT_TIMEOUT,
+            self.timeout_add(5 * 1000, self.send_disconnect, proto, ConnectionMessage.CLIENT_EXIT_TIMEOUT,
                              "version sent")
             return True
+        if request == "id":
+            proto._log_stats = False
+            proto.send_now(("hello", self.get_session_id_info()))
+            self.timeout_add(5 * 1000, self.send_disconnect, proto, ConnectionMessage.CLIENT_EXIT_TIMEOUT,
+                             "id info sent")
+            return True
         return False
 
+    def handle_control(self, command: Sequence[str]) -> bool:
+        log.info(f"handle_control({command})")
+
+        return False
+
+    def get_session_id_info(self) -> dict[str, Any]:
+        # minimal information for identifying the session
+        return {
+            "session-type": "proxy-instance",
+            "uuid": self.uuid,
+            "platform": sys.platform,
+            "pid": os.getpid(),
+            "machine-id": get_machine_id(),
+        }
+
     ################################################################################
 
     def stop(self, skip_proto, *reasons) -> None:
